"""Text related utils."""

import os
import re


def indent(instr, nspaces=4, ntabs=0, flatten=False):
    """Indent a string a given number of spaces or tabstops.

    indent(str,nspaces=4,ntabs=0) -> indent str by ntabs+nspaces.

    Parameters
    ----------

    instr : basestring
        The string to be indented.
    nspaces : int (default: 4)
        The number of spaces to be indented.
    ntabs : int (default: 0)
        The number of tabs to be indented.
    flatten : bool (default: False)
        Whether to scrub existing indentation.  If True, all lines will be
        aligned to the same indentation.  If False, existing indentation will
        be strictly increased.

    Returns
    -------

    str|unicode : string indented by ntabs and nspaces.

    """
    if instr is None:
        return None
    ind = "\t" * ntabs + " " * nspaces
    pat = re.compile("^\\s*", re.MULTILINE) if flatten else re.compile("^", re.MULTILINE)
    outstr = re.sub(pat, ind, instr)
    if outstr.endswith(os.linesep + ind):
        return outstr[: -len(ind)]
    return outstr



"""A preprocessor that extracts all of the outputs from the
notebook file.  The extracted outputs are returned in the 'resources' dictionary.
"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import json
import os
import sys
from binascii import a2b_base64
from mimetypes import guess_extension
from textwrap import dedent

from traitlets import Set, Unicode

from .base import Preprocessor


def guess_extension_without_jpe(mimetype):
    """
    This function fixes a problem with '.jpe' extensions
    of jpeg images which are then not recognised by latex.
    For any other case, the function works in the same way
    as mimetypes.guess_extension
    """
    ext = guess_extension(mimetype)
    if ext == ".jpe":
        ext = ".jpeg"
    return ext


def platform_utf_8_encode(data):
    """Encode data based on platform."""
    if isinstance(data, str):
        if sys.platform == "win32":
            data = data.replace("\n", "\r\n")
        data = data.encode("utf-8")
    return data


class ExtractOutputPreprocessor(Preprocessor):
    """
    Extracts all of the outputs from the notebook file.  The extracted
    outputs are returned in the 'resources' dictionary.
    """

    output_filename_template = Unicode("{unique_key}_{cell_index}_{index}{extension}").tag(
        config=True
    )

    extract_output_types = Set({"image/png", "image/jpeg", "image/svg+xml", "application/pdf"}).tag(
        config=True
    )

    def preprocess_cell(self, cell, resources, cell_index):
        """
        Apply a transformation on each cell,

        Parameters
        ----------
        cell : NotebookNode cell
            Notebook cell being processed
        resources : dictionary
            Additional resources used in the conversion process.  Allows
            preprocessors to pass variables into the Jinja engine.
        cell_index : int
            Index of the cell being processed (see base.py)
        """

        # Get the unique key from the resource dict if it exists.  If it does not
        # exist, use 'output' as the default.  Also, get files directory if it
        # has been specified
        unique_key = resources.get("unique_key", "output")
        output_files_dir = resources.get("output_files_dir", None)

        # Make sure outputs key exists
        if not isinstance(resources["outputs"], dict):
            resources["outputs"] = {}

        # Loop through all of the outputs in the cell
        for index, out in enumerate(cell.get("outputs", [])):
            if out.output_type not in {"display_data", "execute_result"}:
                continue
            if "text/html" in out.data:
                out["data"]["text/html"] = dedent(out["data"]["text/html"])
            # Get the output in data formats that the template needs extracted
            for mime_type in self.extract_output_types:
                if mime_type in out.data:
                    data = out.data[mime_type]

                    # Binary files are base64-encoded, SVG is already XML
                    if mime_type in {"image/png", "image/jpeg", "application/pdf"}:
                        # data is b64-encoded as text (str, unicode),
                        # we want the original bytes
                        data = a2b_base64(data)
                    elif mime_type == "application/json" or not isinstance(data, str):
                        # Data is either JSON-like and was parsed into a Python
                        # object according to the spec, or data is for sure
                        # JSON. In the latter case we want to go extra sure that
                        # we enclose a scalar string value into extra quotes by
                        # serializing it properly.
                        if isinstance(data, bytes):
                            # We need to guess the encoding in this
                            # instance. Some modules that return raw data like
                            # svg can leave the data in byte form instead of str
                            data = data.decode("utf-8")
                        data = platform_utf_8_encode(json.dumps(data))
                    else:
                        # All other text_type data will fall into this path
                        data = platform_utf_8_encode(data)

                    ext = guess_extension_without_jpe(mime_type)
                    if ext is None:
                        ext = "." + mime_type.rsplit("/")[-1]
                    if out.metadata.get("filename", ""):
                        filename = out.metadata["filename"]
                        if not filename.endswith(ext):
                            filename += ext
                    else:
                        filename = self.output_filename_template.format(
                            unique_key=unique_key, cell_index=cell_index, index=index, extension=ext
                        )

                    # On the cell, make the figure available via
                    #   cell.outputs[i].metadata.filenames['mime/type']
                    # where
                    #   cell.outputs[i].data['mime/type'] contains the data
                    if output_files_dir is not None:
                        filename = os.path.join(output_files_dir, filename)
                    out.metadata.setdefault("filenames", {})
                    out.metadata["filenames"][mime_type] = filename

                    if filename in resources["outputs"]:
                        msg = (
                            "Your outputs have filename metadata associated "
                            "with them. Nbconvert saves these outputs to "
                            "external files using this filename metadata. "
                            "Filenames need to be unique across the notebook, "
                            f"or images will be overwritten. The filename {filename} is "
                            "associated with more than one output. The second "
                            "output associated with this filename is in cell "
                            f"{cell_index}."
                        )
                        raise ValueError(msg)
                    # In the resources, make the figure available via
                    #   resources['outputs']['filename'] = data
                    resources["outputs"][filename] = data

        return cell, resources



"""filters for metadata"""


def get_metadata(output, key, mimetype=None):
    """Resolve an output metadata key

    If mimetype given, resolve at mimetype level first,
    then fallback to top-level.
    Otherwise, just resolve at top-level.
    Returns None if no data found.
    """
    md = output.get("metadata") or {}
    if mimetype and mimetype in md:
        value = md[mimetype].get(key)
        if value is not None:
            return value
    return md.get(key)



from nbconvert.utils.text import indent

from .ansi import ansi2html, ansi2latex, strip_ansi
from .citation import citation2latex
from .datatypefilter import DataTypeFilter
from .highlight import Highlight2HTML, Highlight2Latex
from .latex import escape_latex
from .markdown import (
    markdown2asciidoc,
    markdown2html,
    markdown2html_mistune,
    markdown2html_pandoc,
    markdown2latex,
    markdown2rst,
)
from .metadata import get_metadata
from .pandoc import ConvertExplicitlyRelativePaths, convert_pandoc
from .strings import (
    add_anchor,
    add_prompts,
    ascii_only,
    clean_html,
    comment_lines,
    get_lines,
    html2text,
    ipython2python,
    path2url,
    posix_path,
    prevent_list_blocks,
    strip_dollars,
    strip_files_prefix,
    strip_trailing_newline,
    text_base64,
    wrap_text,
)

__all__ = [
    "indent",
    "ansi2html",
    "ansi2latex",
    "strip_ansi",
    "citation2latex",
    "DataTypeFilter",
    "Highlight2HTML",
    "Highlight2Latex",
    "escape_latex",
    "markdown2html",
    "markdown2html_pandoc",
    "markdown2html_mistune",
    "markdown2latex",
    "markdown2rst",
    "markdown2asciidoc",
    "get_metadata",
    "convert_pandoc",
    "ConvertExplicitlyRelativePaths",
    "wrap_text",
    "html2text",
    "clean_html",
    "add_anchor",
    "strip_dollars",
    "strip_files_prefix",
    "comment_lines",
    "get_lines",
    "ipython2python",
    "posix_path",
    "path2url",
    "add_prompts",
    "ascii_only",
    "prevent_list_blocks",
    "strip_trailing_newline",
    "text_base64",
]



"""Generic script exporter class for any kernel language"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import sys

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points  # type:ignore[import-not-found]
else:
    from importlib.metadata import entry_points
from traitlets import Dict, default

from .base import get_exporter
from .templateexporter import TemplateExporter


class ScriptExporter(TemplateExporter):
    """A script exporter."""

    # Caches of already looked-up and instantiated exporters for delegation:
    _exporters = Dict()
    _lang_exporters = Dict()
    export_from_notebook = "Script"

    @default("template_file")
    def _template_file_default(self):
        return "script.j2"

    @default("template_name")
    def _template_name_default(self):
        return "script"

    def _get_language_exporter(self, lang_name):
        """Find an exporter for the language name from notebook metadata.

        Uses the nbconvert.exporters.script group of entry points.
        Returns None if no exporter is found.
        """
        if lang_name not in self._lang_exporters:
            try:
                exporters = entry_points(group="nbconvert.exporters.script")
                exporter = [e for e in exporters if e.name == lang_name][0].load()  # noqa: RUF015
            except (KeyError, IndexError):
                self._lang_exporters[lang_name] = None
            else:
                # TODO: passing config is wrong, but changing this revealed more complicated issues
                self._lang_exporters[lang_name] = exporter(config=self.config, parent=self)
        return self._lang_exporters[lang_name]

    def from_notebook_node(self, nb, resources=None, **kw):
        """Convert from notebook node."""
        langinfo = nb.metadata.get("language_info", {})

        # delegate to custom exporter, if specified
        exporter_name = langinfo.get("nbconvert_exporter")
        if exporter_name and exporter_name != "script":
            self.log.debug("Loading script exporter: %s", exporter_name)
            if exporter_name not in self._exporters:
                exporter = get_exporter(exporter_name)
                # TODO: passing config is wrong, but changing this revealed more complicated issues
                self._exporters[exporter_name] = exporter(config=self.config, parent=self)
            exporter = self._exporters[exporter_name]
            return exporter.from_notebook_node(nb, resources, **kw)

        # Look up a script exporter for this notebook's language
        lang_name = langinfo.get("name")
        if lang_name:
            self.log.debug("Using script exporter for language: %s", lang_name)
            exporter = self._get_language_exporter(lang_name)
            if exporter is not None:
                return exporter.from_notebook_node(nb, resources, **kw)

        # Fall back to plain script export
        self.file_extension = langinfo.get("file_extension", ".txt")
        self.output_mimetype = langinfo.get("mimetype", "text/plain")
        return super().from_notebook_node(nb, resources, **kw)
